#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2024 "Ice Column Cavern" December 28, 2024

in vec3 in_Position;
out vec4 uv;
out float time;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};

#define AA_SAMPLES 4
#define ANIMATION 1
#define SCREENSHOT 2

#define PI 3.141592653589793
#define NO_HIT vec4(-1e8)



//r2 hash http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
vec2 QuasiRandom(float n) {
	#define R2V 1.32471795724474602596
	return fract(.5+n*(1./vec2(R2V,R2V*R2V)));
}

//Panini projection, project screen uv(-1 to 1) to camera, returns ray direction.
vec3 PaniniProjection(vec2 uvp, float fovAngle) {
	vec3 rd = normalize(vec3(uvp,tan(PI*.5-fovAngle*0.25)));//initial planar projection
	vec2 rd2 = normalize(rd.xz);
	float d2 = rd2.y+sqrt(rd2.y*rd2.y);//ray to cylinder xz
	vec2 cd = rd2*d2, cp = cd+vec2(0,-1);
	rd = normalize(vec3(cp.x, rd.y*(dot(cd,cp)/dot(rd.xz,cp)), cp.y));//planar projection onto cylinder
	return rd;
}


//2d rotation matrix from angle
mat2 R2D(float ang) {
	float c = cos(ang), s = sin(ang);
	return mat2(c,s,-s,c);
}

//triangle wave
float TriWave(float p) {
	return abs(fract(p)*2.-1.);
}
vec3 TriWave(vec3 p) {
	return abs(fract(p)*2.-1.);
}


//scene geometry signed distance function
float Geometry(vec3 p) {
	//ice deformation
	p += TriWave(p.zxy*.06)*2.;
	p += TriWave(p.zxy*.11)*1.4;
	p += TriWave(p.zxy*.8428)*.1;
	p += TriWave(p.zxy*2.1942)*.04;
	
	//room domes
	vec3 room = vec3(p.xy, mod(abs(p.z),40.)-20.);
	float d = -min(length(room)-30., length(room.xz)-10.);
	
	//pillars
	p.x += cos(p.z*.1)*4.;
	vec2 pillars = mod(abs(p.xz),10.)-5.;
	d = min(d, length(pillars)-3.+p.y*.08);
	
	//floor
	d = min(d, p.y);
	return d;
}


//ray march scene geometry
vec4 RayMarch(vec3 rayPos, vec3 rayDir, const int steps, float epsilon, const float epsilonExpansion) {
	float dither = .7+.15*fract(abs(dot(rayPos,rayDir.zxy))*1000.);
	for (int i = 0; i < steps; i++) {
		float d = Geometry(rayPos);
		if (d < epsilon) return vec4(rayPos,d);
		rayPos += rayDir*d*dither;
		
		epsilon *= epsilonExpansion;
	}
	return NO_HIT;
}


//ray march inside scene geometry
vec4 RayMarchInside(vec3 rayPos, vec3 rayDir, const int steps, const float epsilon) {
	float tdst = 0.;
	for (int i = 0; i < steps; i++) {
		float d = Geometry(rayPos+rayDir*tdst);
		if (d > epsilon) break;
		tdst += max(abs(d), epsilon);
	}
	return vec4(rayPos+rayDir*tdst, tdst);
}


//scene geometry normal
vec3 Normal(vec4 hit, const float epsilon) {
	vec3 nrm = vec3(Geometry(hit.xyz+vec3(epsilon,0,0)),
					Geometry(hit.xyz+vec3(0,epsilon,0)),
					Geometry(hit.xyz+vec3(0,0,epsilon)))-hit.w;
	float nlen = length(nrm);
	if (nlen == 0.) return vec3(0,1,0);
	return nrm/nlen;
}


//diffuse lighting
vec3 Diffuse(vec3 p, vec3 rd, vec3 nrm) {
	//diffuse light position/direction
	vec3 lpos = vec3(p.x, p.y-20., mod(abs(p.z),40.)-20.);
	float llen = length(lpos);
	lpos *= -1./max(1e-7,llen);
	
	//soft occlusion trace towards light
	float occlusion = 0., s = .1, td = 0.;
	for (int i = 0; i < 8; i++) {
		td += s;
		if (td > llen) break;
		occlusion += clamp(.5-Geometry(p+lpos*td)/s, 0.,1.);
		s *= 1.7;
	}

	occlusion = pow(max(.5, 1.-occlusion*.2), 4.);
	llen = max((1./(1.+llen*.4))*occlusion,
				max(0., (1.-length(lpos.xz))*clamp((p.y+0.)/30.,0.,1.)));
	vec3 diffuse = vec3(.4,.7,.95)*llen*mix(vec3(.001,.04,.2), vec3(1,1,.5)*2., llen);
	
	//inner ice volume shading
	vec3 refr = refract(rd,nrm, .94);
	vec4 refrHit = RayMarchInside(p-nrm*1e-1,refr, 10,1e-1);
	diffuse += pow(vec3(.5,.8,.92), vec3(1.+refrHit.w*.5))*.02;
	
	return diffuse;
}



uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION
	time = float(frame)/float(framerate*AA_SAMPLES);
	#else
	time = 0.;
	#endif
	
	uv = vec4((in_Position.xy*vec2(screenX/screenY,1)+(QuasiRandom(float(frame%AA_SAMPLES))-.5)*2./screenY),
		(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//camera ray
	#ifdef ANIMATION
	float camZ = 200.+tanh((time/60.-.5)*3.)*200.;
	vec3 rp = vec3(cos(camZ*.1)*-4.+TriWave(camZ*.06)*-2., 5, camZ),
		rd = PaniniProjection(uv.xy, PI*1.2);
	#else
	//screenshot 1
	#if SCREENSHOT == 1
	vec3 rp = vec3(-11,8,204),
		rd = PaniniProjection(uv.xy, PI*1.3);
	rd.xz *= R2D(1.3);
	#else
	
	//screenshot 2
	vec3 rp = vec3(19,4,576),
		rd = PaniniProjection(uv.xy, PI*.8);
	rd.yz *= R2D(.4);
	rd.xz *= R2D(-1.38);
	#endif
	#endif
		
	//ray march camera to scene
	vec4 sceneHit = RayMarch(rp,rd, 80,1e-3,1.08);
	vec3 c = vec3(0);
	if (sceneHit.w != NO_HIT) {
		//diffuse shading
		vec3 nrm = Normal(sceneHit, 1e-3);
		c = Diffuse(sceneHit.xyz, rd, nrm);
		
		//specular reflection
		vec3 reflDir = reflect(rd,nrm);
		vec4 reflHit = RayMarch(sceneHit.xyz+nrm*1e-1, reflDir, 50,1e-2,1.1);
		if (reflHit != NO_HIT) {
			//reflection shading
			float fres = pow(max(0.,1.+dot(rd,nrm)),2.);
			c += vec3(.75,.95,1)*fres*Diffuse(reflHit.xyz, reflDir, Normal(reflHit,1e-3));
		}
	}
	
	
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (frame%AA_SAMPLES == 0) sum = vec4(0);
	sum += vec4(c, 1.);
	gl_FragColor = sum;
}
